/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"
#include "uart.h"

struct config config;

static HANDLE hRxUart;

/* http://src.gnu-darwin.org/src/lib/libc/string/strnstr.c.html */
static char *strnstr(const char *s, const char *find, size_t slen)
{
	char c;
	char sc;
	size_t len;

	if ((c = *find++) != '\0') {
		len = strlen(find);
		do {
			do {
				if (slen-- < 1 || (sc = *s++) == '\0')
					return NULL;
			} while (sc != c);
			if (len > slen)
				return NULL;
		} while (strncmp(s, find, len) != 0);
		s--;
	}
	return ((char *)s);
}

int at_receive_line(uint8_t *buf, unsigned int max_length, unsigned int *length, int timeout)
{
	int ret;

	ret = uart_receive_line(buf, max_length, length, timeout);

#if 1 // def DEBUG
	{
		char *ptr = buf;
		int len = *length;
		int res;

		if (ptr[0] == '\r' && ptr[1] == '\n') {
			ptr += 2;
			len -= 2;
		}
		char *ptr2 = ptr;

		const char ev_preamble[] = "+EV ";
		res = memcmp(ptr2, ev_preamble, sizeof(ev_preamble) - 1);
		if (!res) {
			ptr2 += sizeof(ev_preamble) - 1;
		}

		/* "truncate" length to the next "space" */
		char *ptr3 = strnstr(ptr2, " ", len);
		if (ptr3) {
			len = (int)(ptr3 - ptr);
		}
		/* or if that is not found, to the \r\n */
		else {
			ptr3 = strnstr(ptr2, "\r\n", len);
			if (ptr3) {
				len = (int)(ptr3 - ptr);
			}
		}
		
		printf("<-- '%.*s'\n", len, ptr);
	}
#endif

	return ret;
}

uint8_t buf[MAX_BUF_SIZE];

static int at_event(void)
{
	uint8_t *ascii;
	uint8_t *binary;
	int ascii_length;
	int binary_length;
	unsigned int length;
	unsigned int i;
	int ret;

	ret = at_receive_line(buf, sizeof(buf), &length, 100);
	if (ret) {
		dprintf("at_receive_line() failed\n");
		return -1;
	}

	char *ptr = buf;
	i = 0;

	/* skip cr and lf at the beginning */
	for (i = 0; i < length; i++) {
		if (*ptr == '\n' || *ptr == '\r') {
			ptr++;
			continue;
		}
		break;
	}
	if (i == length) {
		dprintf("error parsing event #1\n");
		return -1;
	}

	const char ev_preamble[] = "+EV ";
	ret = memcmp(ptr, ev_preamble, sizeof(ev_preamble)-1);
	if (ret) {
		dprintf("error parsing event #2\n");
		return -1;
	}
	ptr += sizeof(ev_preamble)-1;

	ascii = ptr;
	unsigned int offset = i;

	/* find arguments */
	for (; i < length; i++) {
		if (*ptr != ' ' && *ptr != '=') {
			ptr++;
			continue;
		}
		break;
	}
	if (i == length) {
		dprintf("error parsing event #3\n");
		return -1;
	}
	ascii_length = (int)(ptr - &buf[offset]);
	/* zero terminate command string for sure */
	*ptr++ = 0x0;

	binary = ptr;
	/* find cr and lf at the end */
	for (; i < length - 1; i++) {
		if (*ptr == '\r' && *(ptr+1) == '\n') {
			break;
		}
		ptr++;
	}

	binary_length = (int)(ptr - binary);

	return application_handle_at_event(ascii, ascii_length, binary, binary_length);
}

static int at_send_bytes(const uint8_t *buf, unsigned int length)
{
#if 1 // def DEBUG
	{
		const char *cmd = buf + 3;
		char *ptr = strstr(cmd, " ");
		int len;
		if (ptr) {
			len = (int)(ptr - cmd);
		}
		else {
			len = (int)strlen(cmd) - 2;
		}
		printf("--> '%.*s'\n", len, cmd);
	}
#endif
	return uart_send_bytes(buf, length);
}

static uint8_t status_ok[] = { '+', 'O', 'K' };
static uint8_t status_error[] = { '+', 'E', 'R', 'R', 'O', 'R', '=' };

static int receive_command_status(void)
{
	unsigned int length;
	int ret;

	ret = at_receive_line(buf, sizeof(buf), &length, 1000);
	if (ret) {
		dprintf("at_receive_line() failed\n");
		return -1;
	}

	uint8_t *ptr = buf;

	ret = memcmp(ptr, status_error, sizeof(status_error));
	if (!ret) {
		ptr += sizeof(status_error);
		int error_no = *ptr++;
		printf("+ERROR found, error_no %d\n", error_no);
		return -error_no;
	}

	ret = memcmp(ptr, status_ok, sizeof(status_ok));
	if (!ret) {
		dprintf("+OK found\n");
		return 0;
	}

	printf("unknown response '%s' received\n", buf);
	return -1;
}

int at_send_command(const uint8_t *cmd, unsigned int len)
{
	int ret;

	/* cmd and buf might point to the same memory location.
	copy the cmd first (possibly just moving the data), then add
	the length information. */
	memmove(&buf[3], cmd, len);
	buf[0] = (len + 3) & 0xff;
	buf[1] = ((len + 3) >> 8) & 0xff;
	buf[2] = ((len + 3) >> 16) & 0xff;

	ret = at_send_bytes(buf, len + 3);
	if (ret) {
		dprintf("at_send_bytes() failed\n");
		return -1;
	}

	ret = receive_command_status();
	if (ret) {
		dprintf("receive_command_status() failed, error code %d\n", ret);
		return -ret;
	}

	return 0;
}

int send_at_lediscon(uint8_t *bd_address)
{
	uint8_t command[] = {
		'A', 'T', '+', 'L', 'E', 'D', 'I', 'S', 'C', 'O', 'N', ' ', /* command */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* bd address */
		END
	};
	int ret;

	memcpy(&command[12], bd_address, 6);

	ret = at_send_command(command, sizeof(command));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	dprintf("\n");
	return 0;
}

int main(int argc, char *argv[])
{
	char *device_name = NULL;
	int ret;

	/* use argv[1] as device name */
	if (argc > 1) {
		device_name = argv[1];
	}

	ret = uart_init(&hRxUart, FALSE);
	if (ret) {
		dprintf("uart_init() failed\n");
		return -1;
	}
	
	ret = application_init(device_name);
	if (ret) {
		dprintf("application_inits() failed\n");
		return -1;
	}

	HANDLE hStdin;

	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	if (hStdin == INVALID_HANDLE_VALUE) {
		dprintf("GetStdHandle() failed\n");
		return -1;
	}

	DWORD fdwMode;

	ret = GetConsoleMode(hStdin, &fdwMode);
	if (!ret) {
		dprintf("GetConsoleMode() failed\n");
		return -1;
	}

	fdwMode = fdwMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_MOUSE_INPUT | ENABLE_PROCESSED_INPUT);

	ret = SetConsoleMode(hStdin, fdwMode);
	if (!ret) {
		dprintf("SetConsoleMode() failed\n");
		return -1;
	}

	FlushConsoleInputBuffer(hStdin);
	if (!ret) {
		dprintf("FlushConsoleInputBuffer() failed\n");
		return -1;
	}

	bool shift_pressed = false;
	bool ctrl_pressed = false;

	while (1) {

		/* check if application needs to do any Bluetooth-related processing */
		ret = application_process();
		if (ret) {
			dprintf("application_process() failed\n");
			return -1;
		}

		HANDLE handles[] = { hRxUart, hStdin, };
		DWORD nCount = sizeof(handles) / sizeof(handles[0]);
		ret = WaitForMultipleObjects(nCount, &handles[0], FALSE, INFINITE);

		switch (ret) {
		case WAIT_TIMEOUT: {
			dprintf("WAIT_TIMEOUT @ WaitForMultipleObjects()\n");
			continue;
		}
		case WAIT_FAILED: {
			dprintf("WAIT_FAILED @ WaitForMultipleObjects()\n");
			goto out;
		}
		case WAIT_OBJECT_0: {
			ret = at_event();
			if (ret) {
				dprintf("handle_at_event() failed\n");
				//goto out;
			}
			break;
		}
		case WAIT_OBJECT_0 + 1: {
			INPUT_RECORD record;
			DWORD numRead;

			if (!ReadConsoleInput(hStdin, &record, 1, &numRead)) {
				dprintf("ReadConsoleInput() failed\n");
				goto out;
			}

			if (record.EventType != KEY_EVENT) {
				// don't care about other console events
				continue;
			}

			if (record.Event.KeyEvent.wVirtualKeyCode == VK_SHIFT) {
				if (record.Event.KeyEvent.bKeyDown) {
					shift_pressed = true;
				}
				else {
					shift_pressed = false;
				}
				continue;
			}

			if (record.Event.KeyEvent.wVirtualKeyCode == VK_CONTROL) {
				if (record.Event.KeyEvent.bKeyDown) {
					ctrl_pressed = true;
				}
				else {
					ctrl_pressed = false;
				}
				continue;
			}

			if (!record.Event.KeyEvent.bKeyDown) {
				// really only care about keydown
				continue;
			}

			if (ctrl_pressed == true && record.Event.KeyEvent.wVirtualKeyCode == 0x43) {
				application_shutdown();
				ExitProcess(0);
				continue;
			}

			ret = application_handle_keyboard_input(record.Event.KeyEvent.uChar.AsciiChar);
			if (ret) {
				dprintf("application_handle_keyboard_inputs() failed\n");
			}
		}
		}

	}

	return 0;

out:
	dprintf("Application error\n");
	while (1) {
		Sleep(1000);
	}

	return 0;
}

/* This is a simple ringbuffer implementation. It is mainly used to store any
"keypresses" that are received from the UART for the later forwarding to an
existing Bluetooth connection. */
void flush_rb(struct rb *rb)
{
	memset(rb, 0, sizeof(*rb));
}

int add_to_rb(struct rb *rb, uint8_t c)
{
	ring_pos_t next_head = (rb->ring_head + 1) % RING_SIZE;
	if (next_head != rb->ring_tail) {
		/* there is room */
		rb->ring_data[rb->ring_head] = c;
		rb->ring_head = next_head;
		return 0;
	}

	/* no room left in the buffer */
	return -1;
}

int remove_from_rb(struct rb *rb, uint8_t *c)
{
	ring_pos_t ring_head = rb->ring_head;

	if (ring_head != rb->ring_tail) {
		*c = rb->ring_data[rb->ring_tail];
		rb->ring_tail = (rb->ring_tail + 1) % RING_SIZE;
		return 0;
	}

	return -1;
}
