/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#include "application.h"

static const uint8_t appearance[] = { 0x00, 0x00 };
static uint8_t device_name[0x1e] = "PAN-USB-P-AT";
static size_t device_name_len = 9;
static const uint8_t manufacturer_name_string[] = "Panasonic";
static const uint8_t model_number_string[] = "PAN 1760A";
static const uint8_t system_id[] = "ETU";

static int send_data(const uint8_t *data, int len)
{
	int ret;

	if (config.is_sending == true) {
		dprintf("already sending\n");
		return -1;
	}

	uint16_t timeout = 1000;

	uint8_t at_gssrvind_preamble[] = {
		'A', 'T', '+', 'G', 'S', 'S', 'R', 'V', 'I', 'N', 'D', ' ',
		'P', 3, '+', 'C', 1, COMMA,
		0x00, 0x01, COMMA, /* connection id */
		timeout >> 8, timeout & 0xff, COMMA, /* timeout */
	};

	uint8_t *ptr = buf;
	memcpy(ptr, at_gssrvind_preamble, sizeof(at_gssrvind_preamble));
	ptr += sizeof(at_gssrvind_preamble);
	memcpy(ptr, data, len);
	ptr += len;
	*ptr++ = '\r';
	*ptr++ = '\n';

	dprintf("len %d\n", len);

	ret = at_send_command(buf, (unsigned int)(ptr - buf));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	config.is_sending = true;

	return 0;
}

static int start_advertising(void)
{
	uint8_t command_preamble[] = {
		'A', 'T', '+', 'L', 'E', 'S', 'P', 'P', 'A', 'D', 'V', ' ', /* command */
		0x01, COMMA, /* enable flag */
		/* device name + end is added dynamically below */
	};
	int ret;

	dprintf("\n");

	uint8_t *ptr = buf;
	memcpy(ptr, command_preamble, sizeof(command_preamble));
	ptr += sizeof(command_preamble);
	memcpy(ptr, device_name, device_name_len);
	ptr += device_name_len;
	*ptr++ = '\r';
	*ptr++ = '\n';

	ret = at_send_command(buf, (unsigned int)(ptr - buf));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	return 0;
}

static int stop_advertising(void)
{
	uint8_t command[] = {
		'A', 'T', '+', 'L', 'E', 'S', 'P', 'P', 'A', 'D', 'V', ' ', /* command */
		0x00, /* enable flag */
		END
	};
	int ret;

	dprintf("\n");

	ret = at_send_command(command, sizeof(command));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	return 0;
}

int send_at_gswrupd(uint16_t connection_id)
{
	int ret;

	uint8_t at_gswrupd[] = {
		'A', 'T', '+', 'G', 'S', 'W', 'R', 'U', 'P', 'D', ' ',
		connection_id >> 8, connection_id & 0xff, COMMA, /* connection id */
		0, /* accept */
		END
	};

	dprintf("sending confirmation\n");

	ret = at_send_command(at_gswrupd, sizeof(at_gswrupd));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	return 0;
}

static int retrieve_bluetooth_device_address(uint16_t connection_id, uint8_t *bd_address)
{
	int ret;

	uint8_t at_leconinfo[] = {
		'A', 'T', '+', 'L', 'E', 'C', 'O', 'N', 'I', 'N', 'F', 'O', ' ',
		connection_id >> 8, connection_id & 0xff, /* connection id */
		END
	};

	dprintf("retrieve connection information\n");

	ret = at_send_command(at_leconinfo, sizeof(at_leconinfo));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	unsigned int length;

	ret = at_receive_line(buf, sizeof(buf), &length, 1000);
	if (ret) {
		dprintf("at_receive_line() failed\n");
		return -1;
	}

	uint8_t *ptr = buf;

	static uint8_t leconinfo[] = { '+', 'L', 'E', 'C', 'O', 'N', 'I', 'N', 'F', 'O', '=' };

	ret = memcmp(ptr, leconinfo, sizeof(leconinfo));
	if (ret) {
		dprintf("did not receive leconinfo\n");
		return -1;
	}

	ptr += sizeof(leconinfo);

	int offset = 0;
	uint16_t p_connection_id = (ptr[offset] << 8) | ptr[offset + 1];
	if (p_connection_id != connection_id) {
		dprintf("unexpected connection_id %04x\n", p_connection_id);
	}
	offset += 2;
	offset++; /* comma */
	memcpy(bd_address, &ptr[offset], 6);
	offset += 6;
	/* ignore the rest */

	return 0;
}

static int server_spp_rx_data_buffer_characteristic(uint8_t *data, int len)
{
	dprintf("SERVER_SPP_RX_DATA: buffer characteristic, length %d\n", len);

	printf("'%.*s'\n", len, data);

	return send_at_gswrupd(config.connection_id);
}

static int server_spp_rx_data_device_name_characteristic_handler(uint8_t *data, int len)
{
	dprintf("SERVER_SPP_RX_DATA: device name characteristic, length %d\n", len);

	memcpy(device_name, data, len);
	device_name_len = len;

	return send_at_gswrupd(config.connection_id);
}

static int le_conn_completed_event_handler(uint8_t *buf, int len)
{
	int ret;

	int offset = 0;
	uint8_t status = buf[offset++];
	offset++; /* comma */
	uint16_t connection_id = (buf[offset] << 8) | buf[offset + 1];
	offset += 2;
	offset++; /* comma */
	uint16_t mtu = (buf[offset] << 8) | buf[offset + 1];

	config.mtu_size = mtu;
	config.is_connected = true;
	config.connection_id = connection_id;

	ret = retrieve_bluetooth_device_address(config.connection_id, config.bd_address);
	if (ret) {
		dprintf("retrieve_bluetooth_device_address() failed\n");
		return -1;
	}

	dprintf("LE_CONN_COMPLETED_EVENT: status %d, connection_id %04x, bd_addr "BD_ADDR_FORMAT"\n", status, connection_id, BD_ADDR_ARG(config.bd_address));
	return 0;
}

static int server_write_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler(uint8_t *value)
{
	uint16_t client_characteristic_configuration_descriptor = value[0] | (value[1] << 8);

	dprintf("%04x\n", client_characteristic_configuration_descriptor);

	config.is_indication_enabled = client_characteristic_configuration_descriptor == 0x02 ? true : false;

	return send_at_gswrupd(config.connection_id);
}

static int server_read_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler(void)
{
	uint16_t client_characteristic_configuration_descriptor = 0;
	int ret;

	dprintf("is_indication_enabled %d\n", config.is_indication_enabled);

	if (config.is_indication_enabled) {
		client_characteristic_configuration_descriptor = 0x02;
	}

	uint8_t at_gsrdupd[] = {
		'A', 'T', '+', 'G', 'S', 'R', 'D', 'U', 'P', 'D', ' ',
		'P', 3, '+', 'C', 1, '+', 'E', 2, COMMA,
		config.connection_id >> 8, config.connection_id & 0xff, COMMA, /* connection id */
		1, COMMA, /* update flag */
		0, COMMA, /* accept value */
		client_characteristic_configuration_descriptor & 0xff,
		(client_characteristic_configuration_descriptor >> 8) & 0xff,
		END
	};

	dprintf("sending confirmation\n");

	ret = at_send_command(at_gsrdupd, sizeof(at_gsrdupd));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	return 0;
}

static int send_at_gsrdupd(uint8_t service, uint8_t characteristic, uint8_t element, const uint8_t *data, size_t len)
{
	int ret;

	uint8_t at_gsrdupd_preamble[] = {
		'A', 'T', '+', 'G', 'S', 'R', 'D', 'U', 'P', 'D', ' ',
		'P', service, '+', 'C', characteristic, '+', 'E', element, COMMA,
		config.connection_id >> 8, config.connection_id & 0xff, COMMA, /* connection id */
		1, COMMA, /* update flag */
		0, COMMA, /* accept value */
							/* data + end is missing and is added dynamically */
	};

	uint8_t *ptr = buf;
	memcpy(ptr, at_gsrdupd_preamble, sizeof(at_gsrdupd_preamble));
	ptr += sizeof(at_gsrdupd_preamble);
	memcpy(ptr, data, len);
	ptr += len;
	*ptr++ = '\r';
	*ptr++ = '\n';

	printf("sending data with len %zd for service %d, characteristic %d, element %d\n", len, service, characteristic, element);

	ret = at_send_command(buf, (unsigned int)(ptr - buf));
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	return 0;
}

int application_handle_at_event(uint8_t *ascii, int ascii_length, uint8_t *binary, int binary_len)
{
	int ret;

	ret = strcmp(ascii, "LE_CONN_COMPLETED_EVENT");
	if (!ret) {
		return le_conn_completed_event_handler(binary, binary_len);
	}

	ret = strcmp(ascii, "LE_CONN_UPDATE_EVENT");
	if (!ret) {
		int offset = 0;
		uint8_t status = binary[offset++];
		offset++; /* comma */
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		uint16_t conn_intrv = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		uint16_t conn_lat = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		uint16_t sup_tout = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;

		dprintf("LE_CONN_UPDATE_EVENT: connection_id %04x, conn_intrv %04x, conn_lat %04x, sup_tout %04x\n", connection_id, conn_intrv, conn_lat, sup_tout);

		return 0;
	}

	ret = strcmp(ascii, "LE_DISCONN_EVENT");
	if (!ret) {
		int offset = 0;
		uint8_t reason = binary[offset++];
		offset++; /* comma */
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;

		flush_rb(&config.in);
		memset(&config, 0, sizeof(config));
		dprintf("LE_DISCONN_EVENT: reason %d, connection_id %04x\n", reason, connection_id);

		ret = start_advertising();
		if (ret) {
			dprintf("start_advertising() failed\n");
			return -1;
		}

		return 0;
	}

	ret = strcmp(ascii, "SERVER_MTU_CHANGE_EVENT");
	if (!ret) {
		uint16_t mtu = (binary[3] << 8) | binary[4];
		dprintf("SERVER_MTU_CHANGE_EVENT: mtu %d\n", mtu);
		config.mtu_size = mtu;
		return 0;
	}

	ret = strcmp(ascii, "SERVER_SPP_RX_DATA");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Ci */
		offset++; /* + */
		offset++; /* E */
		uint8_t element = binary[offset++]; /* Ei */
		offset++; /* comma */

		if (service == 1 && characteristic == 1 && element == 1) {
			/* found write to device information characteristic of generic access service */
			return server_spp_rx_data_device_name_characteristic_handler(&binary[offset], binary_len - offset);
		}

		if (service == 3 && characteristic == 1 && element == 1 ) {
			/* found write to buffer characteristic */
			return server_spp_rx_data_buffer_characteristic(&binary[offset], binary_len - offset);
		}

		/* fall through: unknown request */
	}

	ret = strcmp(ascii, "SERVER_READ_LONG_EVENT");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Ci */
		offset++; /* + */
		offset++; /* E */
		uint8_t element = binary[offset++]; /* Ei */
		offset++; /* comma */
		uint16_t value_offset = (binary[offset] << 8) | binary[offset + 1];

		/* fall through - unknown combination of service / characteristic / element */
		dprintf("SERVER_READ_LONG_EVENT: unhandled service %d, characteristic %d, element %d, value_offset %d\n", service, characteristic, element, value_offset);

		uint8_t at_gsrdupd[] = {
			'A', 'T', '+', 'G', 'S', 'R', 'D', 'U', 'P', 'D', ' ',
			'P', service, '+', 'C', characteristic, '+', 'E', element, COMMA,
			config.connection_id >> 8, config.connection_id & 0xff, COMMA, /* connection id */
			0, COMMA, /* update flag */
			0, /* accept value */	
			END
		};

		ret = at_send_command(at_gsrdupd, sizeof(at_gsrdupd));
		if (ret) {
			dprintf("at_send_command() failed\n");
			return -1;
		}

		return 0;
	}

	ret = strcmp(ascii, "SERVER_READ_EVENT");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Ci */
		offset++; /* + */
		offset++; /* E */
		uint8_t element = binary[offset++]; /* Ei */

		if (service == 3 && characteristic == 1 && element == 2) {
			return server_read_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler();
		}

		if (service == 1 && characteristic == 1 && element == 1) {
			return send_at_gsrdupd(1, 1, 1, device_name, device_name_len);
		}

		if (service == 1 && characteristic == 2 && element == 1) {
			return send_at_gsrdupd(1, 2, 1, appearance, sizeof(appearance));
		}

		if (service == 2 && characteristic == 1 && element == 1) {
			return send_at_gsrdupd(2, 1, 1, manufacturer_name_string, sizeof(manufacturer_name_string));
		}

		if (service == 2 && characteristic == 2 && element == 1) {
			return send_at_gsrdupd(2, 2, 1, model_number_string, sizeof(model_number_string));
		}

		if (service == 2 && characteristic == 3 && element == 1) {
			return send_at_gsrdupd(2, 3, 1, system_id, sizeof(system_id));
		}

		/* fall through - unknown combination of service / characteristic / element */
		dprintf("SERVER_READ_EVENT: unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);
	}

	ret = strcmp(ascii, "SERVER_PREP_WRITE_EVENT");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Ci */
		offset++; /* + */
		offset++; /* E */
		uint8_t element = binary[offset++]; /* Ei */
		offset++; /* comma */
		uint16_t write_offset = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;

		if (service == 3 && characteristic == 1 && element == 1) {
			dprintf("SERVER_PREP_WRITE_EVENT: service %d, characteristic %d, element %d, write_offset %04x\n", service, characteristic, element, write_offset);
			return send_at_gswrupd(connection_id);
		}

		/* fall through - unknown combination of service / characteristic / element */
		dprintf("SERVER_PREP_WRITE_EVENT: unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);
	}

	ret = strcmp(ascii, "SERVER_WRITE_EVENT");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Ci */
		offset++; /* + */
		offset++; /* E */
		uint8_t element = binary[offset++]; /* Ei */
		offset++; /* comma */
		uint8_t *value = &binary[offset];

		if (service == 3 && characteristic == 1 && element == 2) {
			return server_write_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler(value);
		}

		/* fall through - unknown combination of service / characteristic / element */
		dprintf("SERVER_WRITE_EVENT: unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);
	}

	ret = strcmp(ascii, "SERVER_INDICATION_ACK_EVENT");
	if (!ret) {

		int offset = 0;
		uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
		offset += 2;
		offset++; /* comma */
		uint8_t status = binary[offset++];
		offset++; /* comma */
		offset++; /* P */
		uint8_t service = binary[offset++]; /* Pi */
		offset++; /* + */
		offset++; /* C */
		uint8_t characteristic = binary[offset++]; /* Cj */

		if (connection_id != config.connection_id) {
			dprintf("unexpected connection_id %04x\n", connection_id);
			return -1;
		}
		if (status) {
			dprintf("unexpected status %02x\n", status);
			return -1;
		}
		if (service != 3) {
			dprintf("unexpected service %02x\n", service);
			return -1;
		}
		if (characteristic != 1) {
			dprintf("unexpected characteristic %02x\n", characteristic);
			return -1;
		}

		config.is_sending = false;

		dprintf("SERVER_INDICATION_ACK_EVENT\n");

		return 0;
	}

	warning("unhandled event '%s'\n", ascii);
	return 0;
}

int application_init(const char *new_device_name)
{
#define DEFAULT_MTU 23

	/* no safety check here */
	if (new_device_name) {
		device_name_len = strlen(new_device_name);
		memcpy(device_name, new_device_name, device_name_len);
	}

	uint8_t init[] = {
		'A', 'T', '+', 'L', 'E', 'S', 'P', 'P', 'I', 'N', 'I', 'T', ' ', /* command */
		0x00, COMMA, /* authentication requirement */
		0x00, COMMA, /* uuid flag */
		appearance[0], appearance[1], COMMA, /* appearance */
		(DEFAULT_MTU >> 8) & 0xff, DEFAULT_MTU & 0xff, COMMA, /* mtu */
		0x00, COMMA, /* connection parameters */
		0x00, /* scan / advertise flag */
		END
	};

	int ret;

	dprintf("\n");

	config.mtu_size = DEFAULT_MTU;

	ret = at_send_command(init, sizeof(init));
	if (ret == 20) {
		printf("Sending AT command failed - Do you need to reset the device?\n");
		return -1;
	}
	if (ret) {
		dprintf("at_send_command() failed\n");
		return -1;
	}

	ret = start_advertising();
	if (ret) {
		dprintf("start_advertising() failed\n");
		return 0;
	}

	return 0;
}

void application_shutdown(void)
{
	int ret;

	if (config.is_connected == false) {
		return;
	}

	ret = send_at_lediscon(config.bd_address);
	if (ret) {
		dprintf("send_at_lediscon() failed\n");
		return;
	}
}

int application_handle_keyboard_input(int key)
{
	if (!config.is_connected) {
		dprintf("not connected - doing nothing\n");
		return 0;
	}

	if (!config.is_indication_enabled) {
		dprintf("indication not enabled - doing nothing\n");
		return 0;
	}

	add_to_rb(&config.in, (uint8_t)key);

	return 0;
}

/* application handling again */
int application_process(void)
{
	int len = 0;
	int ret;

	if (!config.is_connected) {
//		dprintf("not connected - doing nothing\n");
		return 0;
	}

	if (!config.is_indication_enabled) {
//		dprintf("indication not enabled - doing nothing\n");
		return 0;
	}

	if (config.is_sending == true) {
//		dprintf("currently sending - doing nothing\n");
		return 0;
	}

	while (len < config.mtu_size) {
		ret = remove_from_rb(&config.in, &config.buf2[len]);
		if (ret) {
			break;
		}
		len++;
	}

	if (!len) {
		// dprintf("no data to send - doing nothing\n");
		return 0;
	}

	ret = send_data(config.buf2, len);
	if (ret) {
		dprintf("send_data() failed\n");
		return -1;
	}

	config.is_sending = true;

	//	dprintf("length %d\n", length);

	return 0;
}
