/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#include "application.h"

#include <assert.h>

#include "ftd2xx.h"

static FT_HANDLE ftHandle;
static HANDLE rxHandle;

int uart_init(HANDLE *handle, BOOL flow_control)
{
	FT_STATUS ftStatus;
	DWORD offset = 0;
	int open = 0;
	DWORD NumDevs;

	ftStatus = FT_CreateDeviceInfoList(&NumDevs);
	if (ftStatus != FT_OK) {
    dprintf("FT_CreateDeviceInfoList() failed\n");
		return -1;
	}

	if (NumDevs == 0) {
    printf("No devices found\n");
		return -1;
	}

  dprintf("NumDevs %d\n", NumDevs);

	FT_DEVICE_LIST_INFO_NODE *devInfo;

	devInfo = (FT_DEVICE_LIST_INFO_NODE*)malloc(sizeof(FT_DEVICE_LIST_INFO_NODE)*NumDevs);
	if (!devInfo) {
    dprintf("malloc() failed\n");
		return -1;
	}

	ftStatus = FT_GetDeviceInfoList(devInfo, &NumDevs);
	if (ftStatus != FT_OK) {
    dprintf("FT_GetDeviceInfoList() failed\n");
		return -1;
	}

	for (DWORD i = 0; i < NumDevs; i++) {
    dprintf("Dev %d:\n", i);
    dprintf("  Flags=0x%x\n", devInfo[i].Flags);
    dprintf("  Type=0x%x\n", devInfo[i].Type);
    dprintf("  ID=0x%x\n", devInfo[i].ID);
    dprintf("  LocId=0x%x\n", devInfo[i].LocId);
    dprintf("  SerialNumber=%s\n", devInfo[i].SerialNumber);
    dprintf("  Description=%s\n", devInfo[i].Description);
    dprintf("  ftHandle=0x%p\n", devInfo[i].ftHandle);
		if (devInfo[i].Flags & FT_FLAGS_OPENED) {
			open++;
			continue;
		}
		offset = i;
	}

	free(devInfo);

	if (NumDevs - open != 1) {
		printf("Did not find single non-open FTDI device. Aborting.");
		return -1;
	}

	if (offset >= NumDevs) {
    dprintf("offset %d >= NumDevs %d\n", offset, NumDevs);
    return -1;
  }

  ftStatus = FT_Open(offset, &ftHandle);
	if (ftStatus != FT_OK) {
    dprintf("FT_Open() failed\n");
		return -1;
	}

	LONG lComPortNumber;

	ftStatus = FT_GetComPortNumber(ftHandle, &lComPortNumber);
	if (ftStatus != FT_OK) {
    dprintf("FT_GetComPortNumber() failed\n");
		return -1;
	}

	printf("Using device @ COM %d\n", lComPortNumber);

	ftStatus = FT_SetBaudRate(ftHandle, 115200);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetBaudRate() failed\n");
		return -1;
	}

	// Set 8 data bits, 1 stop bit and no parity
	ftStatus = FT_SetDataCharacteristics(ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetDataCharacteristics() failed\n");
		return -1;
	}

	ftStatus = FT_SetFlowControl(ftHandle, flow_control == TRUE ? FT_FLOW_RTS_CTS : FT_FLOW_NONE, 0x11, 0x13);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetFlowControl() failed\n");
		return -1;
	}

	ftStatus = FT_SetTimeouts(ftHandle, 1000, 1000);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetTimeouts() failed\n");
		return -1;
	}

	ftStatus = FT_SetLatencyTimer(ftHandle, 2);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetLatencyTimer() failed\n");
		return -1;
	}

	ftStatus = FT_SetUSBParameters(ftHandle, 512, 512);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetUSBParameters() failed\n");
		return -1;
	}

	ftStatus = FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
	if (ftStatus != FT_OK) {
    dprintf("FT_Purge() failed\n");
		return -1;
	}

	rxHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (!rxHandle) {
    dprintf("CreateEvent() failed\n");
		return -1;
	}
	*handle = rxHandle;

	DWORD EventMask = FT_EVENT_RXCHAR;
	ftStatus = FT_SetEventNotification(ftHandle, EventMask, *handle);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetEventNotification() failed\n");
		return -1;
	}

  dprintf("\n");
	return 0;
}

int uart_set_baudrate(int baudrate)
{
	FT_STATUS ftStatus;

	ftStatus = FT_SetBaudRate(ftHandle, baudrate);
	if (ftStatus != FT_OK) {
    dprintf("FT_SetBaudRate() failed\n");
		return -1;
	}

  dprintf("\n");
  return 0;
}

int uart_send_bytes(const uint8_t *buf, unsigned int length)
{
	FT_STATUS ftStatus;
	DWORD BytesWritten;

#if 0
  {
    unsigned int i;
    printf("%4d --> ", length);
    for (i = 0; i < length; i++)
      printf("0x%02x, ", buf[i]);
    printf("\n");
  }
#endif
  
  ftStatus = FT_Write(ftHandle, (LPVOID)buf, length, &BytesWritten);
	if (ftStatus != FT_OK) {
    dprintf("FT_Write() failed\n");
		return -1;
	}

	assert(length == BytesWritten);

  dprintf("\n");
  return 0;
}

static int receive_bytes(uint8_t *buf, int len, int timeout)
{
  FT_STATUS ftStatus;
  DWORD EventDWord;
  DWORD TxBytes;
  DWORD RxBytes;
  DWORD BytesReceived;
  int ret;

  DWORD todo = len;
  while (todo) {

    ftStatus = FT_GetStatus(ftHandle, &RxBytes, &TxBytes, &EventDWord);
    if (ftStatus != FT_OK) {
      dprintf("FT_GetStatus() failed\n");
      return -1;
    }

    if (!RxBytes) {

      ret = WaitForSingleObject(rxHandle, timeout);
      if (ret == WAIT_TIMEOUT) {
        dprintf("WAIT_TIMEOUT @ WaitForSingleObject()\n");
        return -1;
      }
      if (ret == WAIT_FAILED) {
        dprintf("WAIT_FAILED @ WaitForSingleObject()\n");
        return -1;
      }
      if (ret != WAIT_OBJECT_0) {
        dprintf("error != WAIT_OBJECT_0 @ WaitForSingleObject()\n");
        return -1;
      }
      continue;
    }

    if (RxBytes > todo) {
      RxBytes = todo;
    }

    ftStatus = FT_Read(ftHandle, buf, RxBytes, &BytesReceived);
    if (ftStatus != FT_OK) {
      dprintf("FT_Read() failed\n");
      return -1;
    }
    if (RxBytes != BytesReceived) {
      dprintf("FT_Read() failed @ BytesReceived\n");
      return -1;
    }

    todo -= RxBytes;
    buf += RxBytes;
  }

  return 0;
}

int uart_receive_line(uint8_t *buf, unsigned int max_length, unsigned int *length, int timeout)
{
  uint8_t data[3];
  unsigned int rx_length;
  int ret;

  if (buf == NULL || max_length == 0 || length == NULL) {
    return -1;
  }

  *length = 0;

  ret = receive_bytes(&data[0], 3, timeout);
  if (ret) {
    dprintf("receive_bytes() failed\n");
    return -1;
  }

  rx_length = data[0];
  rx_length |= (data[1] << 8);
  rx_length |= (data[2] << 16);

  if (max_length < rx_length) {
    dprintf("max_length %d out of bounds, need rx_bytes %d\n", max_length, rx_length);
    return -1;
  }

  ret = receive_bytes(buf, rx_length - 3, timeout);
  if (ret) {
    dprintf("receive_bytes() failed\n");
    return -1;
  }

  *length = rx_length - 3;

#if 0
  {
    unsigned int i;
    printf("%4d <-- ", rx_length);
    printf("0x%02x, ", data[0]);
    printf("0x%02x, ", data[1]);
    printf("0x%02x, ", data[2]);
    for (i = 0; i < rx_length-3; i++)
      printf("0x%02x, ", buf[i]);
    printf("\n");
  }
#endif

  dprintf("\n");
  return 0;
}
