/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"
#include "ble_api_master.h"

#include "client_spp_over_ble_api.h"

#define mprintf(...) do { printf("%s(): ", __func__); printf(__VA_ARGS__); } while(0)

struct ble_api_srv_profile *server = NULL;
struct ble_api_cli_profile *client = &spp_over_ble_cli_profile;

struct config
{
	uint8_t buf[SER_ATT_MTU_SIZE_MAX];
	int mtu_size;
	bool is_connected;
	int connection_handle;
	bool is_sending;
	bool is_scanning;
	struct rb in;
	struct {
		uint8_t bd_address[6];
		enum tcu_smp_address_type address_type;
		uint8_t is_name_found;
		uint8_t is_uuid_found;
	} remote;
};

static struct config config;
static struct ble_api_uuid uuid;

static uint8_t *name;
static uint16_t name_len;

static int scan_start(void)
{
	enum ble_api_result result;

	/* if no name is given, we will let the SDK filter for the SPPoverBLE UUID and connect to the
	first device that is found. */
	if (!name) {
		uuid.uuid_type = client_service_spp_over_ble.rom.uuid_type;
		uuid.uuid = client_service_spp_over_ble.rom.uuid;

		result = ble_api_master_start_scan_filter_uuid(&uuid, LE_PASSIVE, ADDRESS_CONFIGURATION_PUBLIC);
		if (result) {
			dprintf("ble_api_master_start_scan_filter_none() failed\n");
			return -1;
		}
	}
	/* otherwise we will just filter for all devices and do the filtering for both the UUID and the name manually later. 
	this is necessary, because UUID and device name may arrive in the advertising data or the scan response separately... */
	else {
		result = ble_api_master_start_scan_filter_none(LE_ACTIVE, ADDRESS_CONFIGURATION_PUBLIC);
		if (result) {
			dprintf("ble_api_master_start_scan_filter_none() failed\n");
			return -1;
		}
	}

	dprintf("\n");
	return 0;
}

static int scan_stop(void)
{
	enum ble_api_result result;

	result = ble_api_master_scan_disable();
	if (result) {
		dprintf("ble_api_master_scan_disable() failed\n");
		return -1;
	}

	dprintf("\n");
	return 0;
}

int application_init(const char *device_name)
{
	enum ble_api_result result;
	int ret;

	if (device_name) {
		size_t len = strlen(device_name);
		if (len >= UINT16_MAX) {
			dprintf("device_name length %zu out of bounds\n", len);
			return -1;
		}
		name_len = (int)len;
		/* no security / safety checks here */
		name = (uint8_t *)malloc(len);
		memcpy(name, device_name, len);
	}

	struct ble_api_scan_connection_parameters config_parameters = {
		0x20, // LE_SCAN_INTERVAL,
		0x20, // LE_SCAN_WINDOW,
		LE_SCAN_FILTER_POLICY,
		LE_SCAN_DUPLICATION_FILTER,
		0x06, // LE_CONNECTION_INTERVAL_MINIMUM,
		80, // LE_CONNECTION_INTERVAL_MAXIMUM,
		LE_CONNECTION_LATENCY,
		100, // LE_SUPERVISION_TIMEOUT,
		LE_MINIMUM_CE_LENGTH,
		LE_MAXIMUM_CE_LENGTH,
		LE_INITIATOR_FILTER_POLICY,
	};

	result = ble_api_configure_scan_connection_parameters(&config_parameters);
	if (result != LE_API_SUCCESS) {
		return -1;
	}
#if 0
	ret = key_store_ble_init();
	if (ret) {
		dprintf("key_store_ble_init() failed\n");
		return -1;
	}
#endif

	ret = scan_start();
	if (ret) {
		dprintf("scan_start() failed\n");
	}

	printf("Device is ready.\n");

	return 0;
}

/* minimal BLE master handling */
enum ble_api_result ble_application_master_scan_enable_response_callback(uint8_t status)
{
	dprintf("status %02x\n", status);

	config.is_scanning = true;

	return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_scan_disable_response_callback(uint8_t status)
{
	enum ble_api_result result;

	dprintf("status %02x\n", status);

	config.is_scanning = false;

	result = ble_api_master_connect(config.remote.bd_address, config.remote.address_type, SMP_ADDRESS_TYPE_PUBLIC_ADDRESS, 5);
	if (result != LE_API_SUCCESS) {
		dprintf("ble_api_master_connect() failed\n");
		return LE_API_ERROR;
	}

	return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_advertise_report_event_callback(struct tcu_mng_le_adv_report_event *event)
{
	struct ble_api_extract_adv_data extract_adv_data_type;
	enum ble_api_result result;
	int ret;

	/* if we have been given a (device) name, this is the name we need to check the found device against.
	only if it matches we will connect to that device. */
	if (name) {

		result = ble_api_extract_uuid_data(LE_LONG_UUID, event->adv_data, event->adv_data_len, &extract_adv_data_type);
		if (result == LE_API_SUCCESS) {
			ret = memcmp(client_service_spp_over_ble.rom.uuid, extract_adv_data_type.data, LE_LONG_UUID);
			if (!ret) {
				config.remote.is_uuid_found = 1;
				config.remote.address_type = event->address_type;
				memcpy(config.remote.bd_address, event->address, 6);
			}
		}

		result = ble_api_extract_device_name(event->adv_data, event->adv_data_len, &extract_adv_data_type);
		if (result == LE_API_SUCCESS) {
			if (extract_adv_data_type.data_len == name_len) {
				ret = memcmp(name, extract_adv_data_type.data, name_len);
				if (!ret) {
					config.remote.is_name_found = 1;
				}
			}
		}

		if (config.remote.is_name_found == 0 || config.remote.is_uuid_found == 0) {
			return LE_API_SUCCESS;
		}
	}

	/* store information for later connect in scan stop handler */
	config.remote.address_type = event->address_type;
	memcpy(config.remote.bd_address, event->address, 6);

	ret = scan_stop();
	if (ret) {
		dprintf("scan_stop() failed\n");
		return LE_API_ERROR;
	}

	dprintf("\n");
	return LE_API_SUCCESS;
}

/* minimal spp over ble client handling */
void client_descriptor_spp_over_ble_enable_indication_done_callback(uint16_t connection_handle)
{
	mprintf("connection_handle %04x -- connected\n", connection_handle);
	config.is_connected = true;
}

void client_descriptor_spp_over_ble_write_done_callback(uint16_t connection_handle)
{
	dprintf("connection_handle %04x\n", connection_handle);
}

void spp_over_ble_indication_callback(uint8_t *value, uint16_t count, uint16_t connection_handle)
{
	mprintf("received '%.*s'\n", count, value);
}

void client_spp_over_ble_write_done_callback(uint16_t connection_handle)
{
	config.is_sending = false;
	dprintf("connection_handle %04x\n", connection_handle);
}

void client_spp_over_ble_disconnected_callback(uint16_t connection_handle)
{
	int ret;

	mprintf("connection_handle %04x\n", connection_handle);
	
	flush_rb(&config.in);
	memset(&config, 0, sizeof(config));

	ret = scan_start();
	if (ret) {
		dprintf("scan_start() failed\n");
		return;
	}
}

void client_spp_over_ble_profile_found_callback(uint16_t connection_handle, uint16_t mtu_size)
{
	int ret;

	config.mtu_size = mtu_size;
	config.connection_handle = connection_handle;

	ret = client_spp_over_ble_service_enable_indication(connection_handle);
	if (ret) {
		dprintf("client_spp_over_ble_service_enable_indication() failed\n");
		return;
	}

	dprintf("mtu_size %d\n", mtu_size);
}

/* application handling again */
int application_process(void)
{
	int len = 0;
	int ret;

	if (!config.is_connected) {
		dprintf("not connected - doing nothing\n");
		return 0;
	}

	if (config.is_sending == true) {
		dprintf("currently sending - doing nothing\n");
		return 0;
	}

	while (len < config.mtu_size) {
		ret = remove_from_rb(&config.in, &config.buf[len]);
		if (ret) {
			break;
		}
		len++;
	}

	if (!len) {
		dprintf("nothing to send - doing nothing\n");
		return 0;
	}

	ret = client_spp_over_ble_service_write_start(config.buf, len, config.connection_handle);
	if (ret) {
		dprintf("client_spp_over_ble_service_write_start() failed\n");
		return -1;
	}

	config.is_sending = true;

//	dprintf("sending data\n");

	return 0;
}

int application_handle_keyboard_input(int key)
{
	if (!config.is_connected) {
		mprintf("not connected - doing nothing\n");
		return 0;
	}

	add_to_rb(&config.in, (uint8_t)key);
	return 0;
}

/* stupid stubbing */
void ble_application_slave_start_advertising_response_callback(uint8_t status)
{
	__debugbreak();
}

void ble_application_slave_disable_advertise_callback(uint8_t status)
{
	__debugbreak();
}
