/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#include <windows.h>
#include <tchar.h>
#include <process.h>

#include "debug_api.h"
#include "uart_api.h"
#include "version_specific.h"
#include "application.h"

#include "ftd2xx.h"

static FT_HANDLE ftHandle;
static HANDLE rxHandle;

int uart_init(HANDLE *handle, BOOL flow_control)
{
	FT_STATUS ftStatus;
	DWORD offset = 0;
	int open = 0;
	DWORD NumDevs;

	ftStatus = FT_CreateDeviceInfoList(&NumDevs);
	if (ftStatus != FT_OK) {
		printf("FT_CreateDeviceInfoList() failed\n");
		return -1;
	}

	if (NumDevs == 0) {
		printf("No devices found\n");
		return -1;
	}

	FT_DEVICE_LIST_INFO_NODE *devInfo;

	devInfo = (FT_DEVICE_LIST_INFO_NODE*)malloc(sizeof(FT_DEVICE_LIST_INFO_NODE)*NumDevs);
	if (!devInfo) {
		printf("malloc() failed\n");
		return -1;
	}

	ftStatus = FT_GetDeviceInfoList(devInfo, &NumDevs);
	if (ftStatus != FT_OK) {
		printf("FT_GetDeviceInfoList() failed\n");
		return -1;
	}

	for (DWORD i = 0; i < NumDevs; i++) {
		dprintf("Dev %d:\n", i);
		dprintf("  Flags=0x%x\n", devInfo[i].Flags);
		dprintf("  Type=0x%x\n", devInfo[i].Type);
		dprintf("  ID=0x%x\n", devInfo[i].ID);
		dprintf("  LocId=0x%x\n", devInfo[i].LocId);
		dprintf("  SerialNumber=%s\n", devInfo[i].SerialNumber);
		dprintf("  Description=%s\n", devInfo[i].Description);
		dprintf("  ftHandle=0x%p\n", devInfo[i].ftHandle);
		if (devInfo[i].Flags & FT_FLAGS_OPENED) {
			open++;
			continue;
		}
		offset = i;
	}

	free(devInfo);

	if (NumDevs - open != 1) {
		printf("Did not find single non-open FTDI device. Aborting.");
		return -1;
	}

	if (offset >= NumDevs) {
		dprintf("offset %d >= NumDevs %d\n", offset, NumDevs);
		return -1;
	}

	ftStatus = FT_Open(offset, &ftHandle);
	if (ftStatus != FT_OK) {
		printf("FT_Open() failed\n");
		return -1;
	}

	LONG lComPortNumber;

	ftStatus = FT_GetComPortNumber(ftHandle, &lComPortNumber);
	if (ftStatus != FT_OK) {
		printf("FT_GetComPortNumber() failed\n");
		return -1;
	}

	printf("Using device @ COM %d\n", lComPortNumber);

	ftStatus = FT_SetBaudRate(ftHandle, 115200);
	if (ftStatus != FT_OK) {
		printf("FT_SetBaudRate() failed\n");
		return -1;
	}

	// Set 8 data bits, 1 stop bit and no parity
	ftStatus = FT_SetDataCharacteristics(ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
	if (ftStatus != FT_OK) {
		printf("FT_SetDataCharacteristics() failed\n");
		return -1;
	}

	ftStatus = FT_SetFlowControl(ftHandle, flow_control == TRUE ? FT_FLOW_RTS_CTS : FT_FLOW_NONE, 0x11, 0x13);
	if (ftStatus != FT_OK) {
		printf("FT_SetFlowControl() failed\n");
		return -1;
	}

	ftStatus = FT_SetTimeouts(ftHandle, 1000, 1000);
	if (ftStatus != FT_OK) {
		printf("FT_SetTimeouts() failed\n");
		return -1;
	}

	ftStatus = FT_SetLatencyTimer(ftHandle, 2);
	if (ftStatus != FT_OK) {
		printf("FT_SetLatencyTimer() failed\n");
		return -1;
	}

	ftStatus = FT_SetUSBParameters(ftHandle, 512, 512);
	if (ftStatus != FT_OK) {
		printf("FT_SetUSBParameters() failed\n");
		return -1;
	}

	ftStatus = FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
	if (ftStatus != FT_OK) {
		printf("FT_Purge() failed\n");
		return -1;
	}

	rxHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (!rxHandle) {
		printf("CreateEvent() failed\n");
		return -1;
	}
	*handle = rxHandle;

	DWORD EventMask = FT_EVENT_RXCHAR;
	ftStatus = FT_SetEventNotification(ftHandle, EventMask, *handle);
	if (ftStatus != FT_OK) {
		printf("FT_SetEventNotification() failed\n");
		return -1;
	}

	return 0;
}

int uart_set_baudrate(int baudrate)
{
	FT_STATUS ftStatus;

	ftStatus = FT_SetBaudRate(ftHandle, baudrate);
	if (ftStatus != FT_OK) {
		printf("FT_SetBaudRate() failed\n");
		return -1;
	}

	return 0;
}

int uart_application_send_bytes(const uint8_t *buf, int length)
{
	FT_STATUS ftStatus;
	DWORD BytesWritten;

	ftStatus = FT_Write(ftHandle, (LPVOID)buf, length, &BytesWritten);
	if (ftStatus != FT_OK) {
		printf("FT_Write() failed\n");
		return -1;
	}

#if 0
	{
		int i;
		printf("%4d --> ", length);
		for (i = 0; i < length; i++)
			printf("0x%02x ", buf[i]);
		printf("\n");
	}
#endif

	assert(length == BytesWritten);

	return 0;
}
	
int uart_application_receive_receive_bytes(uint8_t *buf, uint32_t length, int timeout)
{
	FT_STATUS ftStatus;
	DWORD EventDWord;
	DWORD TxBytes;
	DWORD RxBytes;
	DWORD BytesReceived;
	int offset;
	int ret;
	
	offset = 0;

	ftStatus = FT_GetStatus(ftHandle, &RxBytes, &TxBytes, &EventDWord);
	if (ftStatus != FT_OK) {
		printf("FT_GetStatus() failed\n");
		return -1;
	}

	/* there is enough data availabe, read it in one go */
	if ((int)RxBytes >= length) {

		ftStatus = FT_Read(ftHandle, buf, length, &BytesReceived);
		if (ftStatus != FT_OK) {
			printf("FT_Read() failed\n");
			return -1;
		}

		if (BytesReceived != length) {
			printf("not enough bytes received\n");
			return -1;
		}

		return 0;
	}

	/* there is not enough data availabe */

	/* first read as much as available to empty the rx buffer, if possible */
	if (RxBytes > 0) {
		ftStatus = FT_Read(ftHandle, buf, RxBytes, &BytesReceived);
		if (ftStatus != FT_OK) {
			printf("FT_Read() failed\n");
			return -1;
		}

		offset += BytesReceived;
		length -= BytesReceived;
	}

	while (1) {
		ret = WaitForSingleObject(rxHandle, timeout);
		if (ret == WAIT_TIMEOUT) {
			printf("WAIT_TIMEOUT @ WaitForSingleObject()\n");
			return -1;
		}
		if (ret == WAIT_FAILED) {
			printf("WAIT_FAILED @ WaitForSingleObject()\n");
			return -1;
		}
		if (ret != WAIT_OBJECT_0) {
			printf("error != WAIT_OBJECT_0 @ WaitForSingleObject()\n");
			return -1;
		}

		ftStatus = FT_Read(ftHandle, buf + offset, length, &BytesReceived);
		if (ftStatus != FT_OK) {
			printf("FT_Read() failed\n");
			return -1;
		}

		if (BytesReceived == length) {
			break;
		}

		offset += BytesReceived;
		length -= BytesReceived;
	}

	return 0;
}

int uart_application_check_data_available_callback(void)
{
	FT_STATUS ftStatus;
	DWORD EventDWord;
	DWORD TxBytes;
	DWORD RxBytes;

	ftStatus = FT_GetStatus(ftHandle, &RxBytes, &TxBytes, &EventDWord);
	if (ftStatus != FT_OK) {
		printf("FT_GetStatus() failed\n");
		return -1;
	}

	if (!RxBytes) {
		return 1;
	}

	return 0;
}

void uart_application_data_receive_debug_callback(uint8_t *buf, int length)
{
#if 0
	int i;
	printf("%4d <-- ", length);
	for (i = 0; i < length; i++) {
		printf("0x%02x ", buf[i]);
	}
	printf("\n");
#endif
}
