/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include <windows.h>
#include <process.h>

#include "application.h"

static uint8_t tcu_implementation_buffer[MAX_BUFFER_SIZE_BLE];
uint8_t *uart_buffer = &tcu_implementation_buffer[0];
uint32_t uart_buffer_len = sizeof(tcu_implementation_buffer);
uint8_t *tcu_buffer = &tcu_implementation_buffer[0];
uint32_t tcu_buffer_len = sizeof(tcu_implementation_buffer);

#define mprintf dprintf

static HANDLE hRxUart;

#include "hci_api.h"

static int tcu_init(enum tsb_bluetooth_chip_version *version)
{
	int ret;

	uint8_t bd_addr[TCU_SIZE_BDADDR];
	uint8_t buf[TCU_SIZE_BDADDR];
	int i;

	ret = hci_api_reset();
	if (ret) {
		dprintf("hci_api_reset() failed\n");
		return -1;
	}

	/* give TC3566x to come up again */
	Sleep(100);

#if 0
	ret = hci_api_set_baudrate_tc35667(HCI_API_BAUDRATE_921600);
	if (ret) {
		dprintf("hci_api_set_baudrate_tc35667() failed\n");
		return -1;
	}

	/* reconfigure baud rate to desired rate */
	ret = uart_set_baudrate(921600);
	if (ret) {
		dprintf("uart_set_baudrate() failed\n");
		return -1;
	}

	/* give TC3566x to come up again */
	Sleep(100);
#endif

	ret = hci_api_open_storage_flash(0x00000001);
	if (ret) {
		return -1;
	}

	ret = hci_api_read_data_from_storage_device(0x02, 0x06, bd_addr);
	if (ret) {
		return -1;
	}

	/* byte order in eeprom is reversed to what HCI controller expects */
	for (i = 0; i < 6; i++) {
		buf[5 - i] = bd_addr[i];
	}

	ret = hci_api_write_bd_addr(&buf[0]);
	if (ret) {
		return -1;
	}

	ret = hci_api_read_firmware_version(version);
	if (ret) {
		dprintf("hci_api_read_firmware_version() failed\n");
		return -1;
	}

	ret = hci_api_switch_to_tcu_mode(*version);
	if (ret) {
		dprintf("hci_api_switch_to_tcu_mode() failed\n");
		return -1;
	}

	/* a minimal delay is needed to allow TC356xx to switch to TCU mode */
	Sleep(30);

	printf("Device is now in TCU mode.\n");

	return 0;
}

static bool is_connected;
static struct ble_api_connection connection;

static void shutdown_application(void)
{
	if (is_connected == true) {
		enum ble_api_result result;

		result = ble_api_disconnect(&connection);
		if (result != LE_API_SUCCESS) {
			dprintf("ble_api_disconnect() failed\n");
			/* fall through */
		}
		/* fall through */
	}
}

/* BLE API handling here */
void ble_application_mng_connect_callback(struct ble_api_connection *c)
{
	dprintf("bd_address "BD_ADDR_FORMAT"\n", BD_ADDR_ARG(c->connect_addr));
	connection = *c;
	is_connected = true;
}

void ble_application_mng_disconnect_callback(struct ble_api_connection *connection, uint8_t reason)
{
	is_connected = false;
	dprintf("reason %02x\n", reason);
}

void ble_application_mng_connection_failed_callback(uint8_t status)
{
	dprintf("status %02x\n", status);
}

/* SMP handling */

#include "ble_api_smp.h"

int ble_application_smp_pairing_request_callback(enum smp_device_role role, uint16_t connection_handle)
{
	mprintf("connection_handle %04x, role %02x\n", connection_handle, role);
	return 0;
}

void ble_application_smp_status_callback(enum smp_status status, uint16_t connection_handle)
{
	mprintf("connection_handle %04x, status %02x\n", connection_handle, status);
}

void ble_application_smp_capability_callback(enum tcu_smp_io_capabilities *io_capabilities)
{
	*io_capabilities = SMP_IO_CAP_NO_IP_NO_OP;
	mprintf("\n");
}

/* application handling */

static int handle_uart(void)
{
	uint8_t *buf = uart_buffer;
	int len = uart_buffer_len;
	int ret;

	while (1) {
		ret = tcu_receive_any_message(buf, len, 0);
		if (ret == 1) {
			/* no more messages to process */
			return 0;
		}
		if (ret != 0) {
			mprintf("tcu_receive_any_message() failed\n");
			return -1;
		}

		enum ble_api_result result;
		result = ble_api_event_handler(server, client, buf);
		/* todo: handle return code here? */
	}

	return 0;
}

int main(int argc, char *argv[])
{
	char *device_name = NULL;
	enum tsb_bluetooth_chip_version version;
	enum ble_api_result result;
	int ret;

	/* use argv[1] as device name */
	if (argc > 1) {
		device_name = argv[1];
	}

	ret = uart_init(&hRxUart, FALSE);
	if (ret) {
		dprintf("uart_init() failed\n");
		return -1;
	}

	ret = tcu_init(&version);
	if (ret) {
		mprintf("tcu_init() failed\n");
		return -1;
	}

	result = ble_api_init(version);
	if (result != LE_API_SUCCESS) {
		mprintf("ble_api_init() failed\n");
		return -1;
	}

	ret = application_init(device_name);
	if (ret) {
		mprintf("application_init() failed\n");
		return -1;
	}

	HANDLE hStdin;

	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	if (hStdin == INVALID_HANDLE_VALUE) {
		mprintf("GetStdHandle() failed\n");
		return -1;
	}

	DWORD fdwMode;

	ret = GetConsoleMode(hStdin, &fdwMode);
	if (!ret) {
		mprintf("GetConsoleMode() failed\n");
		return -1;
	}

	fdwMode = fdwMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_MOUSE_INPUT | ENABLE_PROCESSED_INPUT);

	ret = SetConsoleMode(hStdin, fdwMode);
	if (!ret) {
		mprintf("SetConsoleMode() failed\n");
		return -1;
	}

	FlushConsoleInputBuffer(hStdin);
	if (!ret) {
		mprintf("FlushConsoleInputBuffer() failed\n");
		return -1;
	}

	bool shift_pressed = false;
	bool ctrl_pressed = false;

	while (1) {

		/* check if application needs to do any Bluetooth-related processing */
		ret = application_process();
		if (ret) {
			mprintf("application_process() failed\n");
			return -1;
		}

		HANDLE handles[] = { hRxUart, hStdin, };
		DWORD nCount = sizeof(handles) / sizeof(handles[0]);
		ret = WaitForMultipleObjects(nCount, &handles[0], FALSE, INFINITE);

		switch (ret) {
		case WAIT_TIMEOUT: {
			mprintf("WAIT_TIMEOUT @ WaitForMultipleObjects()\n");
			continue;
		}
		case WAIT_FAILED: {
			mprintf("WAIT_FAILED @ WaitForMultipleObjects()\n");
			goto out;
		}
		case WAIT_OBJECT_0: {
			ret = handle_uart();
			if (ret) {
				mprintf("handle_uart() failed\n");
				goto out;
			}
			break;
		}
		case WAIT_OBJECT_0 + 1: {
			INPUT_RECORD record;
			DWORD numRead;

			if (!ReadConsoleInput(hStdin, &record, 1, &numRead)) {
				mprintf("ReadConsoleInput() failed\n");
				goto out;
			}

			if (record.EventType != KEY_EVENT) {
				// don't care about other console events
				continue;
			}

			if (record.Event.KeyEvent.wVirtualKeyCode == VK_SHIFT) {
				if (record.Event.KeyEvent.bKeyDown) {
					shift_pressed = true;
				}
				else {
					shift_pressed = false;
				}
				continue;
			}

			if (record.Event.KeyEvent.wVirtualKeyCode == VK_CONTROL) {
				if (record.Event.KeyEvent.bKeyDown) {
					ctrl_pressed = true;
				}
				else {
					ctrl_pressed = false;
				}
				continue;
			}

			if (!record.Event.KeyEvent.bKeyDown) {
				// really only care about keydown
				continue;
			}

			if (ctrl_pressed == true && record.Event.KeyEvent.wVirtualKeyCode == 0x43) {
				shutdown_application();
				ExitProcess(0);
				continue;
			}

			int key = record.Event.KeyEvent.uChar.AsciiChar;

			if (shift_pressed) {
				key = toupper(key);
			}

			ret = application_handle_keyboard_input(key);
			if (ret) {
				dprintf("application_handle_keyboard_inputs() failed\n");
			}
		}
	}
	}

out:
	printf("Application ended unexpectedly - did you push the reset button?\n");
	return 0;
}

/* This is a simple ringbuffer implementation. It is mainly used to store any
"keypresses" that are received from the UART for the later forwarding to an
existing Bluetooth connection. */
void flush_rb(struct rb *rb)
{
	memset(rb, 0, sizeof(*rb));
}

int add_to_rb(struct rb *rb, uint8_t c)
{
	ring_pos_t next_head = (rb->ring_head + 1) % RING_SIZE;
	if (next_head != rb->ring_tail) {
		/* there is room */
		rb->ring_data[rb->ring_head] = c;
		rb->ring_head = next_head;
		return 0;
	}

	/* no room left in the buffer */
	return -1;
}

int remove_from_rb(struct rb *rb, uint8_t *c)
{
	ring_pos_t ring_head = rb->ring_head;

	if (ring_head != rb->ring_tail) {
		*c = rb->ring_data[rb->ring_tail];
		rb->ring_tail = (rb->ring_tail + 1) % RING_SIZE;
		return 0;
	}

	return -1;
}
