/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"
#include "ble_api_slave.h"

#include "characteristic_gap_appearance.h"
#include "characteristic_body_sensor_location.h"
#include "characteristic_gap_device_name.h"
#include "characteristic_gap_appearance.h"
#include "characteristic_manufacturer_name_string.h"
#include "characteristic_system_id.h"
#include "characteristic_model_number_string.h"
#include "characteristic_gap_device_name.h"
#include "characteristic_body_sensor_location.h"
#include "characteristic_scan_interval_window.h"
#include "characteristic_scan_refresh.h"

#include "server_spp_over_ble_api.h"

#define mprintf(...) do { printf("%s(): ", __func__); printf(__VA_ARGS__); } while(0)

struct ble_api_srv_profile *server = &spp_over_ble_profile;
struct ble_api_cli_profile *client = NULL;

static const uint8_t manufacturer_name[] = "Panasonic";
static const uint8_t model_number[] = "PAN-USB-ETU";
static const uint8_t manufacturer_identifier[] = { 0x01, 0x02, 0x03, 0x04, 0x05 };
static const uint8_t organizationally_unique_identifier[] = { 0x01, 0x02, 0x03 };

struct config
{
	uint8_t buf[SER_ATT_MTU_SIZE_MAX];
	int mtu_size;
	bool is_connected;
	int connection_handle;
	bool is_sending;
	struct rb in;
};

static struct config config;

static uint8_t *name;
static uint16_t name_len;

int application_init(const char *device_name)
{
	enum ble_api_result result;
	int ret;

#if 0
	ret = key_store_ble_init();
	if (ret) {
		dprintf("key_store_ble_init() failed\n");
		return -1;
	}
#endif

	/* gap */
	if (device_name) {
		size_t len = strlen(device_name);
		if (len >= UINT16_MAX) {
			dprintf("device_name length %zu out of bounds\n", len);
			return -1;
		}
		name_len = (int)len;
		/* no security / safety checks here */
		name = (uint8_t *)malloc(len);
		memcpy(name, device_name, len);
		g_characteristic_gap_device_name.length = name_len;
		g_characteristic_gap_device_name.name = name;
	} else	{
		g_characteristic_gap_device_name.name = "PAN-USB-P";
		g_characteristic_gap_device_name.length = sizeof(g_characteristic_gap_device_name.name) - 1;
	}

	/* device information service */
	g_characteristic_model_number_string.model_number = model_number;
	g_characteristic_model_number_string.length = sizeof(model_number) - 1;

	memcpy(&g_characteristic_system_id.manufacturer_identifier[0], &manufacturer_identifier[0], sizeof(g_characteristic_system_id.manufacturer_identifier));
	memcpy(&g_characteristic_system_id.organizationally_unique_identifier[0], &organizationally_unique_identifier[0], sizeof(g_characteristic_system_id.organizationally_unique_identifier));

	g_characteristic_manufacturer_name_string.manufacturer_name = manufacturer_name;
	g_characteristic_manufacturer_name_string.length = sizeof(manufacturer_name) - 1;

	ret = spp_over_ble_profile_init(name, name_len);
	if (ret) {
		dprintf("spp_over_ble_profile_init() failed\n");
		return -1;
	}

	result = ble_api_peripheral_role_start(ADDRESS_CONFIGURATION_PUBLIC);
	if (result != LE_API_SUCCESS) {
		dprintf("ble_api_peripheral_role_start() failed\n");
		return -1;
	}

	printf("Device is now advertising.\n");
	return 0;
}

/* minimal BLE handling */
void ble_application_slave_start_advertising_response_callback(uint8_t status)
{
	dprintf("status %02x\n", status);
}

void ble_application_slave_disable_advertise_callback(uint8_t status)
{
	dprintf("status %02x\n", status);
}

/* minimal spp over ble server handling */
void spp_over_ble_connect_callback(uint8_t *bd_addr, uint16_t connection_handle, uint16_t mtu_size)
{
	memset(&config, 0, sizeof(config));
	config.is_connected = 1;
	config.connection_handle = connection_handle;
	config.mtu_size = mtu_size;

	mprintf("bd_addr "BD_ADDR_FORMAT", mtu_size %d, connection_handle %04x\n", BD_ADDR_ARG(bd_addr), mtu_size, connection_handle);
}

void spp_over_ble_disconnect_callback(uint16_t connection_handle)
{
	enum ble_api_result result;

	mprintf("connection_handle %04x\n", connection_handle);

	result = ble_api_peripheral_role_start(ADDRESS_CONFIGURATION_PUBLIC);
	if (result != LE_API_SUCCESS) {
		dprintf("ble_api_peripheral_role_start() failed\n");
		return;
	}
}

void spp_over_ble_receive_callback(uint8_t *buf, uint16_t length, uint16_t connection_handle)
{
	mprintf("received '%.*s'\n", length, buf);
}

void spp_over_ble_sent_callback(enum spp_over_ble_status status, uint16_t connection_handle)
{
	dprintf("status %d, connection_handle %04x\n", status, connection_handle);

	config.is_sending = false;
}

static int server_spp_over_ble_send_data(uint8_t *buf, uint16_t length)
{
	int ret;

	if (!config.is_connected) {
		dprintf("not connected - cannot send data\n");
		return -1;
	}

	if (config.is_sending) {
		dprintf("already sending - cannot send data\n");
		return -1;
	}

	ret = spp_over_ble_send(buf, length, config.connection_handle);
	if (ret) {
		dprintf("spp_over_ble_send() failed\n");
		return -1;
	}

	config.is_sending = true;

	dprintf("length %d\n", length);

	return 0;
}

/* application handling again */
int application_process(void)
{
	int len = 0;
	int ret;

	if (!config.is_connected) {
		dprintf("not connected - doing nothing\n");
		return 0;
	}

	if (config.is_sending == true) {
		dprintf("currently sending - doing nothing\n");
		return 0;
	}

	while (len < config.mtu_size) {
		ret = remove_from_rb(&config.in, &config.buf[len]);
		if (ret) {
			break;
		}
		len++;
	}

	if (!len) {
		dprintf("no data to send - doing nothing\n");
		return 0;
	}

	ret = spp_over_ble_send(config.buf, len, config.connection_handle);
	if (ret) {
		dprintf("spp_over_ble_send() failed\n");
		return -1;
	}

	config.is_sending = true;

	dprintf("len %d\n", len);
	return 0;
}

int application_handle_keyboard_input(int key)
{
	if (!config.is_connected) {
		mprintf("not connected - doing nothing\n");
		return 0;
	}

	add_to_rb(&config.in, (uint8_t)key);
	return 0;
}

#if 0
#include "server_heart_rate_api.h"

static struct ble_api_srv_profile *server = &heart_rate_profile;
static struct ble_api_cli_profile *client = NULL;

struct heart_rate_measurement heart_rate_measurement;

int server_profile_init(void)
{
	struct heart_rate_measurement *hr = &heart_rate_measurement;
	int ret;

	memset(hr, 0, sizeof(*hr));
	hr->flags = 0;
	hr->heart_rate_value = 0x55;
	hr->energy_expended_field = 0x0001;

	ret = heart_rate_profile_init(APPEARANCE_CATEGORY_GENERIC_HEART_RATE_SENSOR, BODY_SENSOR_LOCATION_BODY_SENSOR_LOCATION_CHEST);
	if (ret) {
		dprintf("heart_rate_profile_init() failed\n");
		return -1;
	}

	return 0;
}

#endif

/* stupid stubbing */
enum ble_api_result ble_application_master_advertise_report_event_callback(struct tcu_mng_le_adv_report_event *event)
{
	__debugbreak();
	return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_scan_enable_response_callback(uint8_t status)
{
	__debugbreak();
	return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_scan_disable_response_callback(uint8_t status)
{
	__debugbreak();
	return LE_API_SUCCESS;
}
