/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#include "application.h"
#include "debug.h"
#include "system.h"
#include "message.h"
#include "host_api.h"
#include "patch_api.h"
#include "uart1_function.h"

struct ble_api_srv_profile *server = NULL;
struct ble_api_cli_profile *client = NULL;

extern void __iar_data_init3(void);

/* This function reads the pre-programmed Bluetooth device address from the
built-in flash. Surprisingly, this also initializes the Bluetooth core with 
this address. So this procedure is mandatory to be executed. */
static int init_bd_address(void)
{
#define BD_ADDR_LEN 0x06
#define BD_ADDR_OFFSET 0x02
  uint8_t bd_address[BD_ADDR_LEN];
  int ret;
  
  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }
  
  ret = SYS_API_QSPIStorageReadSpecifiedAddr(BD_ADDR_LEN, BD_ADDR_OFFSET);
  if (ret) {
    dprintf("SYS_API_QSPIStorageReadSpecifiedAddr() failed\n");
    return -1;
  }
  
  ret = SYS_API_QSPIStorageRead(BD_ADDR_OFFSET, BD_ADDR_LEN, &bd_address[0]);
  if (ret) {
    dprintf("SYS_API_QSPIStorageRead() failed\n");
    return -1;
  }
  
  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }
  
  dprintf("\n");
  return 0;
}

/* The message subsystem calls this function whenever a TCU message needs to be
processed. Any TCU message is simply forwarded to the BLE API here. */
int message_tcu_command_callback(uint8_t *buf)
{
  enum ble_api_result result;
  
  result = ble_api_event_handler(server, client, buf);
  if (result) {
    dprintf("ble_api_event_handler() failed\n");
    return -1;
  }
  
  dprintf("\n");
  return 0;
}

/* This is the main function. It does the basic initialization and then goes
into the main message processing loop. The "system" and "message" handling 
functions come straight from the Bluetooth SDK. */
#pragma default_function_attributes = @"ENTRY"
void main(void)
{
  enum tsb_bluetooth_chip_version version;
  enum ble_api_result result;
  int ret;
  
  __iar_data_init3();
  
  initializeUart1();
  
  ret = init_bd_address();
  if (ret) {
    dprintf("init_bd_address() failed\n");
    goto fatal_error;
  }
  
  ret = system_read_version(&version);
  if (ret) {
    dprintf("system_read_version() failed\n");
    goto fatal_error;
  }
  
  update_tc35678_79_firmware(version);
  
  result = ble_api_init(version);
  if (result) {
    dprintf("ble_api_init() failed\n");
    goto fatal_error;
  }
  
  /* The peripheral and central demo need different init functions */
  ret = application_init();
  if (ret) {
    dprintf("application_init() failed\n");
    goto fatal_error;
  }
  
  while (1) {
    message_handler();
  }
  
fatal_error:
  warning("fatal error - application did not start\n");
  while (1) {
    /* fixme */
  }
}
#pragma default_function_attributes =

static uint8_t buf[128];
void uart1_printf(const char *format, ...)
{
  va_list args;
  int ret;
  
  va_start(args, format);
  ret = vsnprintf((char *)buf, sizeof(buf), format, args);
  va_end(args);
  
  sendUart1(buf, ret);
}

/* This is a simple ringbuffer implementation. It is mainly used to store any 
"keypresses" that are received from the UART for the later forwarding to an
existing Bluetooth connection. */
void flush_rb(struct rb *rb)
{
  memset(rb, 0, sizeof(*rb));
}

int add_to_rb(struct rb *rb, uint8_t c) 
{
  ring_pos_t next_head = (rb->ring_head + 1) % RING_SIZE;
  if (next_head != rb->ring_tail) {
    /* there is room */
    rb->ring_data[rb->ring_head] = c;
    rb->ring_head = next_head;
    return 0;
  }
  
  /* no room left in the buffer */
  return -1;
}

int remove_from_rb(struct rb *rb, uint8_t *c) 
{
  ring_pos_t ring_head = rb->ring_head;
  
  if (ring_head != rb->ring_tail) {
    *c = rb->ring_data[rb->ring_tail];
    rb->ring_tail = (rb->ring_tail + 1) % RING_SIZE;
    return 0;
  }
  
  return -1;
}

/* This example does not implement the BLE key store API. Some callback functions need
to be stubbed so that the BLE API accepts this fact. */
int key_store_ble_is_irk_available(void)
{
  return 0;
}

int key_store_ble_get_info(uint8_t *bd_address, struct tcu_le_smp_slv_mas_key_info *request)
{
  return -1;
}


int key_store_ble_store_auxiliary_data(uint8_t *bd_address, uint8_t *data, int length)
{
  return -1;
}

void key_store_ble_disconnection_done(uint8_t *bd_address)
{
}

/* Some BLE API connection management functions need to be stubbed. */
void ble_application_mng_connect_callback(struct ble_api_connection *connection)
{
}

void ble_application_mng_disconnect_callback(struct ble_api_connection *connection, uint8_t reason)
{
}

void ble_application_mng_connection_failed_callback(uint8_t status)
{
  /* todo: maybe we need to start advertising / scanning here again */
  uart1_printf("status %02x\n", status);
}

int ble_application_mng_remote_connection_parameter_request_callback(uint16_t connection_handle)
{
  return 0;
}

void ble_application_mng_data_length_change_event_callback(uint8_t status, uint16_t connection_handle, struct le_data_length_event *event)
{
}

