/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#include "application.h"
#include "uart1_function.h"
#include "debug.h"

#include "ble_api_slave.h"
#include "server_spp_over_ble_api.h"
#include "characteristic_gap_device_name.h"
#include "characteristic_gap_appearance.h"
#include "characteristic_manufacturer_name_string.h"
#include "characteristic_system_id.h"
#include "characteristic_model_number_string.h"
#include "characteristic_gap_device_name.h"

/* The configuration of the "device" is set here. */
static const uint8_t manufacturer_name[] = "Panasonic";
static const int manufacturer_name_len = sizeof(manufacturer_name) - 1;

static const uint8_t model_number[] = "Modelnumber";
static const int model_number_len = sizeof(model_number) - 1;

static const uint8_t manufacturer_identifier[] = {0x01, 0x02, 0x03, 0x04, 0x05 };
static const uint8_t organizationally_unique_identifier[] = { 0x01, 0x02, 0x03 };

/* This structure is used to hold all information needed to run the SPPoverBLE
profile. */
static struct { 
  /*! This buffer is used to hold the data that is about to be sent using the
  indication. */
  uint8_t buf[SER_ATT_MTU_SIZE_MAX];
  uint16_t mtu_size;
  uint16_t connection_handle;
  uint8_t is_sending;
  uint8_t is_connected;
  struct rb in;
} config;

static int start_advertising(void)
{
  enum ble_api_result result;
  
  result = ble_api_peripheral_role_start(ADDRESS_CONFIGURATION_PUBLIC);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_peripheral_role_start() failed\n");
    return -1;
  }
  
  dprintf("\n");
  return 0;
}

int application_init(void)
{
  int ret;
  
  /* gap service */
  g_characteristic_gap_device_name.length = device_name_len;
  g_characteristic_gap_device_name.name = device_name;
  
  /* device information service */
  g_characteristic_model_number_string.model_number = model_number;
  g_characteristic_model_number_string.length = model_number_len;
  
  memcpy(&g_characteristic_system_id.manufacturer_identifier[0], &manufacturer_identifier[0], sizeof(g_characteristic_system_id.manufacturer_identifier));
  memcpy(&g_characteristic_system_id.organizationally_unique_identifier[0], &organizationally_unique_identifier[0], sizeof(g_characteristic_system_id.organizationally_unique_identifier));
  
  g_characteristic_manufacturer_name_string.manufacturer_name = manufacturer_name;
  g_characteristic_manufacturer_name_string.length = manufacturer_name_len;
  
  server = &spp_over_ble_profile;
  
  ret = spp_over_ble_profile_init(device_name, device_name_len);
  if (ret) {
    dprintf("server init failed\n");
    return -1;
  }
  
  ret = start_advertising();
  if (ret) {
    dprintf("start_advertising() failed\n");
    return -1;
  }
  
  dprintf("\n");
  aprintf("PERIPHERAL\n");
  return 0;
}

/* Callback function to indicate disable advertise response is received */
void ble_application_slave_disable_advertise_callback(uint8_t status)
{
  dprintf("status 0x%02x\n", status);
}

void ble_application_slave_set_adv_data_response_callback(uint8_t status)
{
  dprintf("status 0x%02x\n", status);
}

void ble_application_slave_start_advertising_response_callback(uint8_t status)
{
  dprintf("status 0x%02x\n", status);
}

void spp_over_ble_connect_callback(uint8_t *bd_addr, uint16_t connection_handle, uint16_t max_packet_size)
{
  config.connection_handle = connection_handle;
  config.mtu_size = max_packet_size - 3;
  /* Only print message once */
  if (!config.is_connected) {
    aprintf("CONNECT\n");
  }
  config.is_connected = 1;
  
  dprintf("bd_addr " BD_ADDR_FORMAT ", max_packet_size %d, connection_handle 0x%04x\n", BD_ADDR_ARG(bd_addr), max_packet_size, connection_handle);
}

void spp_over_ble_disconnect_callback(uint16_t connection_handle)
{
  int ret;
  
  config.is_connected = 0;
  
  ret = start_advertising();
  if (ret) {
    dprintf("start_advertising() failed\n");
    return;
  }
  
  dprintf("connection_handle 0x%04x\n", connection_handle);
  aprintf("DISCONNECT\n");
}

int uart1_receive_bytes_callback_isr(uint8_t *ptr, int len)
{
  int i;
  int ret;
  
  if (!config.is_connected) {
    /* not connected, ignore any input */
    return 0;
  }
  
  for (i = 0; i < len; i++) {
    ret = add_to_rb(&config.in, *ptr);
    if (ret) {
      /* no debug print here, because we are in an isr... */
      //      dprintf("add_to_rb() failed\n");
      return -1;
    }
    ptr++;
  }
  
  return 0;
}

static int pump_bytes_from_uart_to_spp_over_ble(void)
{
  int len = 0;
  int ret;
  
  if (!config.is_connected) {
    dprintf("not connected\n");
    /* not connected, ignore any input */
    return 0;
  }
  
  if (config.is_sending) {
    dprintf("is sending\n");
    /* sending, come back later */
    return 0;
  }
  
  /* len is used to count the number of bytes we are about to send. The
  amount of data depends on the MTU size of the connection. */
  while(len < config.mtu_size) {
    ret = remove_from_rb(&config.in, &config.buf[len]);
    if (ret) {
      break;
    }
    len++;
  }
  
  if (!len) {
    /* nothing to send, to nothing */
    return 0;
  }
  
  ret = spp_over_ble_send(config.buf, len, config.connection_handle);
  if (ret) {
    dprintf("client_spp_over_ble_service_write_start() failed\n");
    return -1;
  }
  
  config.is_sending = 1;
  dprintf("len %d\n", len);
  return 0;
}

void message_uart_event_callback(void)
{
  int ret;
  
  ret = pump_bytes_from_uart_to_spp_over_ble();
  if (ret) {
    dprintf("pump_bytes_from_uart_to_spp_over_ble() failed\n");
    return;
  }
  
  dprintf("\n");
}

int handle_uart1_api_mod(void *msg)
{
  message_uart_event_callback();
  return 0;
}

void spp_over_ble_receive_callback(uint8_t *buf, uint16_t length, uint16_t connection_handle)
{
  sendUart1(buf, length);
  
  dprintf("connection_handle 0x%04x, length %d\n", connection_handle, length);
}

void spp_over_ble_sent_callback(enum spp_over_ble_status status, uint16_t connection_handle)
{
  dprintf("connection_handle 0x%04x, status %d\n", connection_handle, status);
  
  config.is_sending = 0;
  
  /* there may be data left in the buffer, so sent it out now by faking this callback */
  message_uart_event_callback();
}

int auto_speed_test(uint16_t connection_handle)
{
  dprintf("connection_handle 0x%04x\n", connection_handle);
  return 0;
}
